/* Copyright (C) <2011> Luo Jinghua<sunmoon1997@gmail.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <EGL/egl.h>
#include <GLES/gl.h>

#include <sdl-freetype.h>
#include <sdl-freetype-opengles.h>

#ifdef HAVE_FONTCONFIG
#include <sdl-ft.h>
#endif

#include <stdio.h>

static EGLDisplay eglDisplay = 0;
static EGLConfig eglConfig = 0;
static EGLContext eglContext = 0;
static EGLSurface eglSurface = 0;

#ifdef HAVE_X
static Display *x11Display = NULL;
static Window x11Window = None;
#endif

// consts
#define COLOURDEPTH_RED_SIZE  		8
#define COLOURDEPTH_GREEN_SIZE 		8
#define COLOURDEPTH_BLUE_SIZE 		8
#define COLOURDEPTH_DEPTH_SIZE		16

static const EGLint configAttribs[] ={
    EGL_RED_SIZE,      	    COLOURDEPTH_RED_SIZE,
    EGL_GREEN_SIZE,    	    COLOURDEPTH_GREEN_SIZE,
    EGL_BLUE_SIZE,     	    COLOURDEPTH_BLUE_SIZE,
    EGL_DEPTH_SIZE,	    COLOURDEPTH_DEPTH_SIZE,
    EGL_SURFACE_TYPE,       EGL_WINDOW_BIT,
    EGL_RENDERABLE_TYPE,    EGL_OPENGL_ES_BIT,
    EGL_NONE
};

static int InitOpenGL(int width, int height)
{
    EGLint numConfigsOut = 0;
    /* EGLint contextParams[] = {EGL_CONTEXT_CLIENT_VERSION, 1, EGL_NONE};		// Use GLES version 1.x */

#ifdef HAVE_X
    // use EGL to initialise GLES
    x11Display = XOpenDisplay(NULL);

    if (!x11Display)
    {
	fprintf(stderr, "ERROR: unable to get display!n");
	return 0;
    }

    x11Window = XCreateSimpleWindow(x11Display,
				    RootWindow(x11Display, DefaultScreen(x11Display)),
				    1, 1, width, height, 0,
				    BlackPixel(x11Display, DefaultScreen(x11Display)),
				    BlackPixel(x11Display, DefaultScreen(x11Display)));
    if (x11Window == None)
    {
	fprintf(stderr, "ERROR: unable to create a window!n");
	return 0;
    }

    XMapWindow(x11Display, x11Window);
    XFlush(x11Display);

    eglDisplay = eglGetDisplay((EGLNativeDisplayType)x11Display);
#else
    eglDisplay = eglGetDisplay((EGLNativeDisplayType)EGL_DEFAULT_DISPLAY);
#endif
    if (eglDisplay == EGL_NO_DISPLAY)
    {
	printf("Unable to initialise EGL display.");
	return 0;
    }

    // Initialise egl
    if (!eglInitialize(eglDisplay, NULL, NULL))
    {
	printf("Unable to initialise EGL display.");
	return 0;
    }

    // Find a matching config
    if (eglChooseConfig(eglDisplay, configAttribs, &eglConfig, 1, &numConfigsOut) != EGL_TRUE || numConfigsOut == 0)
    {
	fprintf(stderr, "Unable to find appropriate EGL config.");
	return 0;
    }

#ifdef HAVE_X
    eglSurface = eglCreateWindowSurface(eglDisplay, eglConfig, (EGLNativeWindowType)x11Window, 0);
#else
    eglSurface = eglCreateWindowSurface(eglDisplay, eglConfig, (EGLNativeWindowType)0, 0);
#endif
    if ( eglSurface == EGL_NO_SURFACE)
    {
	printf("Unable to create EGL surface!");
	return 0;
    }

    // Bind GLES and create the context
    eglBindAPI(EGL_OPENGL_ES_API);
    eglContext = eglCreateContext(eglDisplay, eglConfig, NULL, NULL);
    if (eglContext == EGL_NO_CONTEXT)
    {
	printf("Unable to create GLES context!");
	return 0;
    }

    if (eglMakeCurrent(eglDisplay,  eglSurface,  eglSurface, eglContext) == EGL_FALSE)
    {
	printf("Unable to make GLES context current");
	return 0;
    }

    return 1;
}

/*======================================================
 * Kill off any opengl specific details
 ====================================================*/
static void TerminateOpenGL(void)
{
    eglMakeCurrent(eglDisplay, NULL, NULL, EGL_NO_CONTEXT);
    eglDestroySurface(eglDisplay, eglSurface);
    eglDestroyContext(eglDisplay, eglContext);
    eglSurface = 0;
    eglContext = 0;
    eglConfig = 0;

    eglTerminate(eglDisplay);
    eglDisplay = 0;

#ifdef HAVE_X
    XDestroyWindow(x11Display, x11Window);
    XCloseDisplay(x11Display);
    x11Display = NULL;
#endif
}

static int SwapBuffers(void)
{
    eglSwapBuffers(eglDisplay, eglSurface);
    return 0;
}

static void DrawLine(float x0, float y0, float x1, float y1)
{
    GLfloat verts[4];

    verts[0] = x0;
    verts[1] = y0;
    verts[2] = x1;
    verts[3] = y1;

    glEnableClientState (GL_VERTEX_ARRAY);

    glVertexPointer (2, GL_FLOAT, 0, verts);
    glDrawArrays (GL_LINES, 0, 2);

    glDisableClientState (GL_VERTEX_ARRAY);
}

#define TEXT "sdl freetype OpenGLES render"
#define WIDTH 640
#define HEIGHT 480
int
main(int argc, char *argv[])
{
    int done;
    sdl_freetype_font_t * font;
    sdl_freetype_glyph_render_t * render;
    sdl_freetype_text_extents_t extents;

    if (!InitOpenGL(WIDTH, HEIGHT))
	return -1;

#ifdef HAVE_FONTCONFIG
    font = sdl_ft_create_font ("sans", 12,
                               SDL_FT_WEIGHT_NORMAL, SDL_FT_SLANT_NORMAL, 96);
#else
    if (argc < 2)
	return -1;
    font = sdl_freetype_font_create(argv[1], 12 * 96 / 72.0);
#endif
    if (!font)
        goto errquit0;

    render = sdl_freetype_opengles_render_create ();
    if (!render)
        goto errquit1;
    sdl_freetype_font_set_render (font, render);

    glMatrixMode (GL_PROJECTION) ;
    glLoadIdentity ();
    glOrthof (0, WIDTH, HEIGHT, 0, -1.0, 1.0);

    glMatrixMode (GL_MODELVIEW);
    glLoadIdentity ();

    done = 0;
    while (!done) {
	float x, y;

#ifdef HAVE_X
	XEvent xevent;

	while (XPending(x11Display)) {
	    XNextEvent(x11Display, &xevent);
	    if (xevent.type == KeyPress)
		done = 1;
	}
#endif

	glClearColor (0.8f, 0.8f, 0.8f, 0.0f);
        glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        sdl_freetype_font_utf8_extents (font, &extents, TEXT, -1);
	x = (WIDTH - extents.width) / 2 - extents.x_bearing;
	y = (HEIGHT - extents.height) / 2 + extents.y_bearing;
        sdl_freetype_font_show_utf8 (font, NULL, 255, 0, 0, 255,
				     x, y, TEXT, -1);
	DrawLine(x, y, x + extents.width, y);

        SwapBuffers();
    }

 errquit1:
    sdl_freetype_font_destroy (font);
    sdl_freetype_fini ();

 errquit0:
    TerminateOpenGL();
    return 0;
}
