/* sdl-freetype, a text rendering library for SDL.
 * Copyright (C) 2006 luojinghua <sunmoon1997@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it either under the terms of the GNU Lesser General Public
 * License version 2.1 as published by the Free Software Foundation
 * (the "LGPL") or, at your option, under the terms of the Mozilla
 * Public License Version 1.1 (the "MPL"). If you do not alter this
 * notice, a recipient may use your version of this file under either
 * the MPL or the LGPL.
 *
 * You should have received a copy of the LGPL along with this library
 * in the file COPYING-LGPL-2.1; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * You should have received a copy of the MPL along with this library
 * in the file COPYING-MPL-1.1
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY
 * OF ANY KIND, either express or implied. See the LGPL or the MPL for
 * the specific language governing rights and limitations.
 *
 * THANKS FOR ALL CAIRO AUTHORS, MOST OF IDEAS AND IMPLEMENT DTAILS ARE
 * STOLEN FROM CAIRO.
 *
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "hash.h"
#include "sdl-fontmap.h"
#include <stdlib.h>
#include <math.h>
#include <assert.h>

struct _sdl_freetype_font_map {
    sdl_freetype_hash_t * hash;
    FT_Library library;
    int num_faces;
    int max_faces;
};

static sdl_freetype_font_map_t fontmap = { NULL, NULL, 0 };

static unsigned long
hash_string(const char * c)
{
    /* This is the djb2 hash. */
    unsigned long hash = 5381;
    while (c && *c)
        hash = ((hash << 5) + hash) + *c++;
    return hash;
}

static unsigned long
fontmap_entry_hash (const sdl_freetype_hash_entry_t * entry)
{
    const sdl_freetype_base_font_t * base_font = 
        (const sdl_freetype_base_font_t *) entry;

    if (base_font->from_memory)
        return (unsigned long)entry->key;
    return hash_string (entry->key) ^ (base_font->index * 3171);
}

static int
fontmap_entry_equal (const sdl_freetype_hash_entry_t * entry0,
                     const sdl_freetype_hash_entry_t * entry1)
{
    const sdl_freetype_base_font_t * base_font0 = 
        (const sdl_freetype_base_font_t *) entry0;
    const sdl_freetype_base_font_t * base_font1 = 
        (const sdl_freetype_base_font_t *) entry1;

    if (base_font0->from_memory && base_font1->from_memory &&
        entry0->key ==  entry1->key &&
        base_font0->index == base_font1->index)
        return 1;
    else if (!base_font0->from_memory && !base_font1->from_memory &&
             !strcmp (entry0->key, entry1->key) &&
             base_font0->index == base_font1->index)
        return 1;

    return 0;
}

static void
fontmap_entry_destroy (sdl_freetype_hash_entry_t * entry)
{
    sdl_freetype_base_font_t * base_font = 
        (sdl_freetype_base_font_t *) entry;

    sdl_freetype_base_font_destroy (base_font);
}

static const sdl_freetype_hash_ops_t fontmap_hash_ops = {
    fontmap_entry_hash, fontmap_entry_equal, fontmap_entry_destroy
};

sdl_freetype_font_map_t *
sdl_freetype_font_map_lock (void)
{
    if (fontmap.library)
        return &fontmap;

    if (FT_Init_FreeType (&fontmap.library))
        goto errquit0;

    fontmap.hash = sdl_freetype_hash_create (10, &fontmap_hash_ops);
    if (!fontmap.hash)
        goto errquit1;
    fontmap.num_faces = 0;
    fontmap.max_faces = 10;

    return &fontmap;
 errquit1:
    FT_Done_FreeType (fontmap.library);
    fontmap.library = NULL;
 errquit0:
    return NULL;
}

void
sdl_freetype_font_map_unlock (sdl_freetype_font_map_t * fontmap)
{
}

static void
sdl_freetype_font_map_faces_destroy_callback (sdl_freetype_hash_t * hash,
                                              sdl_freetype_hash_entry_t * entry,
                                              void * closure)
{
    sdl_freetype_base_font_t * base_font = (sdl_freetype_base_font_t *)entry;
    sdl_freetype_font_face_t * cur, * next;

    cur = base_font->faces;
    base_font->faces = NULL;
    for (; cur; cur = next) {
        next = cur->next;
        sdl_freetype_font_face_destroy (cur);
    }
}

static void
sdl_freetype_font_map_destroy_font (sdl_freetype_font_map_t * fontmap,
                                    sdl_freetype_base_font_t * base_font)
{
    assert (base_font->refcnt == 1 + base_font->num_faces);

    sdl_freetype_font_map_faces_destroy_callback (fontmap->hash, 
                                                  &base_font->base, NULL);
    if (base_font->face) {
        FT_Done_Face (base_font->face);
        base_font->face = NULL;
    }
    base_font->selected = 0;
    base_font->num_faces = 0;
    fontmap->num_faces--;
    sdl_freetype_hash_remove (fontmap->hash, &base_font->base);
}

static void
sdl_freetype_font_map_destroy_faces (sdl_freetype_font_map_t * fontmap)
{
    sdl_freetype_hash_foreach (fontmap->hash,
                               sdl_freetype_font_map_faces_destroy_callback,
                               NULL);
}

static void
sdl_freetype_font_map_destroy (void)
{
    if (!fontmap.library)
        return;

    sdl_freetype_font_map_destroy_faces (&fontmap);
    sdl_freetype_hash_destroy (fontmap.hash);
    fontmap.hash = NULL;
    fontmap.num_faces = 0;

    FT_Done_FreeType (fontmap.library);
    fontmap.library = NULL;
}

static void
sdl_freetype_base_font_key_init (sdl_freetype_base_font_t * base_font,
                                 const char * file, int from_memory,
                                 int size, int index)
{
    base_font->base.key = (void *)file;
    base_font->index = index;
    base_font->from_memory = from_memory;
    base_font->size = size;
}

static int
sdl_freetype_base_font_init (sdl_freetype_base_font_t * base_font,
                             const void * key, int from_memory,
                             off_t size, int index)
{
    memset(base_font, 0, sizeof(sdl_freetype_base_font_t));

    if (from_memory) {
        base_font->base.key = (void *)key;
        base_font->size = size;
    } else {
        base_font->base.key = strdup(key);
        base_font->size = 0;
    }
    base_font->from_memory = from_memory;
    base_font->index = index;
    base_font->refcnt = 1;
    base_font->selected = 0;
    base_font->stream = NULL;
    base_font->num_faces = 0;
    base_font->faces = NULL;

    if (!base_font->base.key)
        return -1;
    return 0;
}

static void
sdl_freetype_base_font_fini (sdl_freetype_base_font_t * base_font)
{
    if (!base_font->from_memory)
        free (base_font->base.key);
    if (base_font->stream)
        free (base_font->stream);
    base_font->stream = NULL;
    base_font->base.key = NULL;
}

static sdl_freetype_base_font_t *
sdl_freetype_base_font_create_intl (const char * file,
                                    int from_memory, off_t size,
                                    int index)
{
    sdl_freetype_base_font_t base_font, * font;
    sdl_freetype_font_map_t * fontmap;

    fontmap = sdl_freetype_font_map_lock ();
    if (!fontmap)
        goto errquit0;
    sdl_freetype_base_font_key_init (&base_font, file, from_memory,
                                     size, index);
    font = (sdl_freetype_base_font_t *)
        sdl_freetype_hash_lookup (fontmap->hash, &base_font.base);
    if (font) {
        sdl_freetype_base_font_reference (font);
    } else {
        font = malloc (sizeof(sdl_freetype_base_font_t));
        if (!font)
            goto errquit1;

        if (sdl_freetype_base_font_init (font, file, from_memory, size,
                                         index) < 0)
            goto errquit2;
        if (sdl_freetype_hash_insert (fontmap->hash, &font->base) < 0)
            goto errquit3;
        sdl_freetype_base_font_reference (font);
    }
    sdl_freetype_font_map_unlock (fontmap);
    return font;
 errquit3:
    sdl_freetype_base_font_fini (font);
 errquit2:
    free (font);
 errquit1:
    sdl_freetype_font_map_unlock (fontmap);
 errquit0:
    return NULL;
}

sdl_freetype_base_font_t *
sdl_freetype_base_font_create (const char * file,
                               int index)
{
    return sdl_freetype_base_font_create_intl (file, 0, 0, index);
}

sdl_freetype_base_font_t *
sdl_freetype_base_font_create_from_memory (void * data, off_t size, int index)
{
    return sdl_freetype_base_font_create_intl (data, 1, size, index);
}

sdl_freetype_base_font_t *
sdl_freetype_base_font_reference (sdl_freetype_base_font_t * base_font)
{
    if (!base_font)
        return base_font;

    base_font->refcnt++;
    return base_font;
}

static int
fontmap_unlocked_filter (sdl_freetype_hash_t * hash,
                         sdl_freetype_hash_entry_t * entry)
{
    sdl_freetype_base_font_t * font = (sdl_freetype_base_font_t *) entry;

    return !font->lock && font->face;
}

static unsigned long 
sdl_freetype_memory_read (FT_Stream stream,
                          unsigned long offset,
                          unsigned char * buffer,
                          unsigned long count)
{
    char * data;

    data = (char *)stream->descriptor.pointer;
    assert (data != NULL);
    if (offset > stream->size)
        return 0;
    if (offset + count > stream->size)
        count = stream->size - offset;
    memcpy (buffer, data + offset, count);
    return count;
}

FT_Face
sdl_freetype_base_font_lock_face (sdl_freetype_base_font_t * base_font)
{
    sdl_freetype_font_map_t * fontmap;

    assert (base_font->lock >= 0);

    if (base_font->face) {
        base_font->lock++;
        return base_font->face;
    }

    fontmap = sdl_freetype_font_map_lock ();
    if (!fontmap)
        goto errquit0;

    while (fontmap->num_faces > fontmap->max_faces) {
        sdl_freetype_base_font_t * font;

        font = (sdl_freetype_base_font_t *)
            sdl_freetype_hash_any_entry (fontmap->hash, fontmap_unlocked_filter);

        if (font && font->face) {
            FT_Done_Face (font->face);
            font->face = NULL;
            font->selected = 0;
            font->num_faces = 0;
            fontmap->num_faces--;
        }
    }

    if (base_font->from_memory) {
        FT_Open_Args args;

        if (!base_font->stream) {
            base_font->stream = calloc(sizeof(*base_font->stream), 1);
            if (!base_font->stream)
                goto errquit1;

            base_font->stream->descriptor.pointer = (char *)base_font->base.key;
            base_font->stream->read = sdl_freetype_memory_read;
            base_font->stream->pos = 0;
            base_font->stream->size = base_font->size;
        }

        args.flags = FT_OPEN_STREAM;
        args.stream = base_font->stream;
        if (FT_Open_Face(fontmap->library, &args,
                         base_font->index, &base_font->face) != FT_Err_Ok)
            goto errquit1;
    } else {
        if (FT_New_Face (fontmap->library, base_font->base.key, base_font->index,
                         &base_font->face) != FT_Err_Ok)
            goto errquit1;
    }

    base_font->lock++;
    fontmap->num_faces++;

    sdl_freetype_font_map_unlock (fontmap);
    return base_font->face;
 errquit1:
    sdl_freetype_font_map_unlock (fontmap);
 errquit0:
    return NULL;
}

static int
ft_matrix_equal (const FT_Matrix * m0,
                 const FT_Matrix * m1)
{
    return m0->xx == m1->xx && m0->xy == m1->xy &&
        m0->yx == m1->yx && m0->yy == m1->yy;
}

void
sdl_freetype_base_font_select_size (sdl_freetype_base_font_t * base_font,
                                    double size, FT_Matrix * xform)
{
    FT_Error error;
    FT_Face face;
    FT_Matrix matrix;

    if (base_font->selected && base_font->size == size &&
        ft_matrix_equal (&base_font->xform, xform))
        return;

    face = base_font->face;
    assert (face != NULL);

    base_font->selected = 1;
    base_font->size = size;
    base_font->xform = *xform;

    matrix.xx = xform->xx;
    matrix.xy = -xform->xy;
    matrix.yx = -xform->yx;
    matrix.yy = xform->yy;

    FT_Set_Transform(base_font->face, &matrix, NULL);

    if (FT_IS_SCALABLE(base_font->face)) {
        error = FT_Set_Char_Size (face,
                                  size * 64.0,
                                  size * 64.0,
                                  0, 0);
        assert (error == 0);
    } else {
        double min_distance = -1;
        int i;
        int best = 0;

        for (i = 0; i < face->num_fixed_sizes; i++) {
#if HAVE_FT_BITMAP_SIZE_Y_PPEM
            double fsize = face->available_sizes[i].y_ppem / 64.;
#else
            double fsize = face->available_sizes[i].height;
#endif
            double distance = fabs (fsize - size);

            if (distance <= min_distance || min_distance < 0) {
                min_distance = distance;
                best = i;
            }
        }
#if HAVE_FT_BITMAP_SIZE_Y_PPEM
        error = FT_Set_Char_Size (face,
                                  face->available_sizes[best].x_ppem,
                                  face->available_sizes[best].y_ppem,
                                  0, 0);
        if (error)
#endif
            error = FT_Set_Pixel_Sizes (face,
                                        face->available_sizes[best].width,
                                        face->available_sizes[best].height);
        assert (error == 0);
    }
}

void
sdl_freetype_base_font_unlock_face (sdl_freetype_base_font_t * base_font)
{
    assert (base_font->lock > 0);
    base_font->lock--;
}

FT_Library
sdl_freetype_base_font_get_freetype_library (sdl_freetype_base_font_t * base_font)
{
    sdl_freetype_font_map_t * fontmap;
    FT_Library library;

    fontmap = sdl_freetype_font_map_lock ();
    if (!fontmap)
        goto errquit0;
    library = fontmap->library;
    sdl_freetype_font_map_unlock (fontmap);
    return library;
 errquit0:
    return NULL;
}

static int
sdl_freetype_font_options_equal (const sdl_freetype_font_options_t * options0,
                                 const sdl_freetype_font_options_t * options1)
{
    if (options0->antialias != options1->antialias)
        return 0;
    if (options0->subpixel_order != options1->subpixel_order)
        return 0;
    if (options0->hint != options1->hint)
        return 0;
    if (options0->embolden != options1->embolden)
        return 0;
    if (options0->embedded_bitmap != options1->embedded_bitmap)
        return 0;
    if (options0->vertical_layout != options1->vertical_layout)
        return 0;
    if (options0->autohint != options1->autohint)
        return 0;
    return 1;
}

static sdl_freetype_font_face_t *
sdl_freetype_base_font_find_face (sdl_freetype_base_font_t * base_font,
                                  const sdl_freetype_font_options_t *options)
{
    sdl_freetype_font_face_t * face;

    for (face = base_font->faces; face; face = face->next)
        if (sdl_freetype_font_options_equal (&face->options, options))
            return sdl_freetype_font_face_reference (face);

    return NULL;
}

static void
sdl_freetype_base_font_append_face (sdl_freetype_base_font_t * base_font,
                                    sdl_freetype_font_face_t * face)
{
    base_font->num_faces++;
    sdl_freetype_font_face_reference (face);
    face->next = base_font->faces;
    base_font->faces = face;
}

static void
sdl_freetype_font_map_destroy_font (sdl_freetype_font_map_t * fontmap,
                                    sdl_freetype_base_font_t * base_font);

void
sdl_freetype_base_font_destroy (sdl_freetype_base_font_t * base_font)
{
    if (!base_font)
        return;

    base_font->refcnt--;
    if (!base_font->refcnt) {
        sdl_freetype_base_font_fini (base_font);
        free (base_font);
    } else if (base_font->from_memory &&
               base_font->refcnt == 1 + base_font->num_faces) {
        sdl_freetype_font_map_t * fontmap;

        fontmap = sdl_freetype_font_map_lock ();
        sdl_freetype_font_map_destroy_font (fontmap, base_font);
        sdl_freetype_font_map_unlock (fontmap);
    }
}

sdl_freetype_font_face_t *
sdl_freetype_font_face_create (sdl_freetype_base_font_t * base_font,
                               const sdl_freetype_font_options_t * options)
{
    sdl_freetype_font_face_t * face;

    face = sdl_freetype_base_font_find_face (base_font, options);
    if (face)
        return face;

    face = malloc (sizeof(sdl_freetype_font_face_t));
    if (!face)
        goto errquit0;

    face->base_font = sdl_freetype_base_font_reference (base_font);
    face->options = *options;
    face->refcnt = 1;

    sdl_freetype_base_font_append_face (base_font, face);
    return face;
 errquit0:
    return NULL;
}

sdl_freetype_font_face_t *
sdl_freetype_font_face_reference (sdl_freetype_font_face_t * face)
{
    if (!face)
        return NULL;

    face->refcnt++;
    return face;
}

void
sdl_freetype_font_face_destroy (sdl_freetype_font_face_t * face)
{
    if (!face)
        return;

    face->refcnt--;
    if (!face->refcnt) {
        sdl_freetype_font_face_t * cur, * prev;
        sdl_freetype_base_font_t * base_font = face->base_font;

        for (cur = base_font->faces, prev = NULL; cur;
             prev = cur, cur = cur->next) {
            if (cur == face) {
                if (!prev)
                    base_font->faces = cur->next;
                else
                    prev->next = cur->next;
            }
        }
        sdl_freetype_base_font_destroy (base_font);

        free (face);
    }
}

int
sdl_freetype_init (void)
{
    sdl_freetype_font_map_t * fontmap;

    fontmap = sdl_freetype_font_map_lock ();
    if (!fontmap)
        return -1;

    sdl_freetype_font_map_unlock (fontmap);
    return 0;
}

void
sdl_freetype_fini (void)
{
    sdl_freetype_font_map_destroy ();
}

