/* sdl-freetype, a text rendering library for SDL.
 * Copyright (C) 2007 luojinghua <sunmoon1997@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it either under the terms of the GNU Lesser General Public
 * License version 2.1 as published by the Free Software Foundation
 * (the "LGPL") or, at your option, under the terms of the Mozilla
 * Public License Version 1.1 (the "MPL"). If you do not alter this
 * notice, a recipient may use your version of this file under either
 * the MPL or the LGPL.
 *
 * You should have received a copy of the LGPL along with this library
 * in the file COPYING-LGPL-2.1; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * You should have received a copy of the MPL along with this library
 * in the file COPYING-MPL-1.1
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY
 * OF ANY KIND, either express or implied. See the LGPL or the MPL for
 * the specific language governing rights and limitations.
 *
 * THANKS FOR ALL CAIRO AUTHORS, MOST OF IDEAS AND IMPLEMENT DTAILS ARE
 * STOLEN FROM CAIRO.
 *
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "sdl-freetype.h"
#include "sdl-freetype-error.h"
#include "sdl-freetype-area-manager.h"
#include <stdlib.h>

struct sdl_freetype_area_manager_s {
    int max_level;
    int width, height;
    sdl_freetype_area_t * area;
    const sdl_freetype_area_manager_ops_t * ops;
};

static sdl_freetype_error_t
sdl_freetype_area_move_in (sdl_freetype_area_manager_t * manager,
                           sdl_freetype_area_t         * area,
                           void                        * closure)
{
    area->closure = closure;
    area->state   = SDL_FREETYPE_AREA_OCCUPIED;

    return manager->ops->move_in (manager, area, area->closure);
}

static void
sdl_freetype_area_move_out (sdl_freetype_area_manager_t * manager,
                            sdl_freetype_area_t         * area)
{
    manager->ops->move_out (manager, area, area->closure);

    area->closure = NULL;
    area->state   = SDL_FREETYPE_AREA_AVAILABLE;
}

void
sdl_freetype_area_manager_move_out (sdl_freetype_area_manager_t * manager,
                                    sdl_freetype_area_t         * area)
{
    if (manager && area)
        sdl_freetype_area_move_out (manager, area);
}

static sdl_freetype_area_t *
sdl_freetype_area_create (sdl_freetype_area_manager_t * manager,
                          int                           level,
                          int                           x,
                          int                           y,
                          int                           width,
                          int                           height)
{
    sdl_freetype_area_t * area;
    int                n = 4;

    area = malloc (sizeof (sdl_freetype_area_t));
    if (!area)
        return NULL;

    area->level   = level;
    area->x       = x;
    area->y       = y;
    area->width   = width;
    area->height  = height;
    area->closure = NULL;
    area->state   = SDL_FREETYPE_AREA_AVAILABLE;

    while (n--)
        area->area[n] = NULL;

    return area;
}

static void
sdl_freetype_area_destroy (sdl_freetype_area_manager_t * manager,
                           sdl_freetype_area_t         * area)
{
    if (!area)
        return;

    if (area->state == SDL_FREETYPE_AREA_OCCUPIED) {
        sdl_freetype_area_move_out (manager, area);
    } else {
        int n = 4;

        while (n--)
            sdl_freetype_area_destroy (manager, area->area[n]);
    }

    free (area);
}

static sdl_freetype_area_t *
sdl_freetype_area_get_top_scored_sub_area (sdl_freetype_area_manager_t * manager,
                                           sdl_freetype_area_t         * area)
{
    if (!area)
        return NULL;

    switch (area->state) {
    case SDL_FREETYPE_AREA_OCCUPIED:
        return area;
    case SDL_FREETYPE_AREA_AVAILABLE:
        break;
    case SDL_FREETYPE_AREA_DIVIDED: {
        sdl_freetype_area_t * tmp, * top = NULL;
        int                   i;

        for (i = 0; i < 4; i++) {
            tmp = sdl_freetype_area_get_top_scored_sub_area (manager, 
                                                             area->area[i]);
            if (tmp && top) {
                if (manager->ops->compare_score (manager,
                                                 tmp,
                                                 tmp->closure,
                                                 top->closure) > 0)
                    top = tmp;
            } else if (tmp) {
                top = tmp;
            }
        }
        return top;
    }
    }

    return NULL;
}

static sdl_freetype_error_t
sdl_freetype_area_manager_find_intl (sdl_freetype_area_manager_t * manager,
                                     sdl_freetype_area_t         * area,
                                     int                           width,
                                     int                           height,
                                     sdl_freetype_bool_t           kick_out,
                                     void                        * closure)
{
    sdl_freetype_error_t err;

    if (area->width < width || area->height < height)
        return SDL_FREETYPE_ERR_OUT_OF_RESOURCES;

    switch (area->state) {
    case SDL_FREETYPE_AREA_OCCUPIED:
        if (kick_out) {
            if (manager->ops->compare_score (manager,
                                             area,
                                             area->closure,
                                             closure) >= 0)
                return SDL_FREETYPE_ERR_OUT_OF_RESOURCES;

            sdl_freetype_area_move_out (manager, area);
        } else {
            return SDL_FREETYPE_ERR_OUT_OF_RESOURCES;
        }

        /* fall-through */
    case SDL_FREETYPE_AREA_AVAILABLE: {
        if (area->level == manager->max_level ||
            (area->width == width && area->height == height)) {
            return sdl_freetype_area_move_in (manager, area, closure);
        } else {
            int dx[4], dy[4], w[4], h[4], i;

            dx[0] = dx[2] = dy[0] = dy[1] = 0;

            w[0] = w[2] = dx[1] = dx[3] = width;
            h[0] = h[1] = dy[2] = dy[3] = height;

            w[1] = w[3] = area->width - width;
            h[2] = h[3] = area->height - height;

            for (i = 0; i < 2; i++) {
                if (w[i])
                    area->area[i] =
                        sdl_freetype_area_create (manager,
                                                  area->level + 1,
                                                  area->x + dx[i],
                                                  area->y + dy[i],
                                                  w[i], h[i]);
            }

            for (; i < 4; i++) {
                if (w[i] && h[i])
                    area->area[i] =
                        sdl_freetype_area_create (manager,
                                                  area->level + 1,
                                                  area->x + dx[i],
                                                  area->y + dy[i],
                                                  w[i], h[i]);
            }

            area->state = SDL_FREETYPE_AREA_DIVIDED;

            err = sdl_freetype_area_manager_find_intl (manager,
                                                       area->area[0],
                                                       width, height,
                                                       kick_out, closure);
            if (err == SDL_FREETYPE_ERR_OK)
                return SDL_FREETYPE_ERR_OK;
        }
    } break;
    case SDL_FREETYPE_AREA_DIVIDED: {
        sdl_freetype_area_t * to_area;
        sdl_freetype_bool_t   rejected = SDL_FREETYPE_FALSE;
        int                   i;

        for (i = 0; i < 4; i++) {
            if (area->area[i]) {
                if (area->area[i]->width >= width &&
                    area->area[i]->height >= height) {
                    err = sdl_freetype_area_manager_find_intl (manager,
                                                               area->area[i],
                                                               width, height,
                                                               kick_out, closure);
                    if (err == SDL_FREETYPE_ERR_OK)
                        return SDL_FREETYPE_ERR_OK;

                    rejected = SDL_FREETYPE_TRUE;
                }
            }
        }

        if (rejected)
            return SDL_FREETYPE_ERR_OUT_OF_RESOURCES;

        to_area = sdl_freetype_area_get_top_scored_sub_area (manager, area);
        if (to_area) {
            if (!kick_out)
                return SDL_FREETYPE_ERR_OUT_OF_RESOURCES;

            if (manager->ops->compare_score (manager,
                                             to_area,
                                             to_area->closure,
                                             closure) >= 0)
                return SDL_FREETYPE_ERR_OUT_OF_RESOURCES;
        }

        for (i = 0; i < 4; i++) {
            sdl_freetype_area_destroy (manager, area->area[i]);
            area->area[i] = NULL;
        }

        area->closure = NULL;
        area->state   = SDL_FREETYPE_AREA_AVAILABLE;

        err = sdl_freetype_area_manager_find_intl (manager, area, width, height,
                                                   SDL_FREETYPE_TRUE, closure);
        if (err == SDL_FREETYPE_ERR_OK)
            return SDL_FREETYPE_ERR_OK;
    } break;
    }

    return SDL_FREETYPE_ERR_OUT_OF_RESOURCES;
}

sdl_freetype_error_t
sdl_freetype_area_manager_find (sdl_freetype_area_manager_t * manager,
                                int                           width,
                                int                           height,
                                sdl_freetype_bool_t           kick_out,
                                void                        * closure)
{
    return sdl_freetype_area_manager_find_intl (manager, manager->area,
                                                width, height, kick_out, closure);
}

static sdl_freetype_error_t
sdl_freetype_area_manager_init (sdl_freetype_area_manager_t           * manager,
                                int                                     width,
                                int                                     height,
                                int                                     max_level,
                                const sdl_freetype_area_manager_ops_t * ops)
{
    manager->max_level = max_level;
    manager->ops       = ops;
    manager->width     = width;
    manager->height    = height;

    manager->area = sdl_freetype_area_create (manager, 0, 0, 0, width, height);
    if (!manager->area)
        return SDL_FREETYPE_ERR_OUT_OF_MEMORY;

    return SDL_FREETYPE_ERR_OK;
}

static void
sdl_freetype_area_manager_fini (sdl_freetype_area_manager_t * manager)
{
    sdl_freetype_area_destroy (manager, manager->area);
}

sdl_freetype_area_manager_t *
sdl_freetype_area_manager_create (int width, int height, int max_level,
                                  const sdl_freetype_area_manager_ops_t * ops)
{
    sdl_freetype_area_manager_t * manager;

    manager = malloc (sizeof (sdl_freetype_area_manager_t));
    if (!manager)
        goto errquit0;
    if (sdl_freetype_area_manager_init (manager, width, height, max_level, ops))
        goto errquit1;

    return manager;
 errquit1:
    free (manager);
 errquit0:
    return NULL;
}

void
sdl_freetype_area_manager_destroy(sdl_freetype_area_manager_t * manager)
{
    if (!manager)
        return;

    sdl_freetype_area_manager_fini (manager);
    free (manager);
}
