/* sdl-freetype, a text rendering library for SDL.
 * Copyright (C) 2006 luojinghua <sunmoon1997@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it either under the terms of the GNU Lesser General Public
 * License version 2.1 as published by the Free Software Foundation
 * (the "LGPL") or, at your option, under the terms of the Mozilla
 * Public License Version 1.1 (the "MPL"). If you do not alter this
 * notice, a recipient may use your version of this file under either
 * the MPL or the LGPL.
 *
 * You should have received a copy of the LGPL along with this library
 * in the file COPYING-LGPL-2.1; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * You should have received a copy of the MPL along with this library
 * in the file COPYING-MPL-1.1
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY
 * OF ANY KIND, either express or implied. See the LGPL or the MPL for
 * the specific language governing rights and limitations.
 *
 * THANKS FOR ALL CAIRO AUTHORS, MOST OF IDEAS AND IMPLEMENT DTAILS ARE
 * STOLEN FROM CAIRO.
 *
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if   HAVE_STDINT_H
# include <stdint.h>
#elif HAVE_INTTYPES_H
# include <inttypes.h>
#elif HAVE_SYS_INT_TYPES_H
# include <sys/int_types.h>
#elif defined(_MSC_VER)
  typedef __int8 int8_t;
  typedef unsigned __int8 uint8_t;
  typedef __int16 int16_t;
  typedef unsigned __int16 uint16_t;
  typedef __int32 int32_t;
  typedef unsigned __int32 uint32_t;
  typedef __int64 int64_t;
  typedef unsigned __int64 uint64_t;
# ifndef HAVE_UINT64_T
#  define HAVE_UINT64_T 1
# endif
# ifndef INT16_MIN
#  define INT16_MIN     (-32767-1)
# endif
# ifndef INT16_MAX
#  define INT16_MAX     (32767)
# endif
# ifndef UINT16_MAX
#  define UINT16_MAX    (65535)
# endif
#else
#error Cannot find definitions for fixed-width integral types (uint8_t, uint32_t, etc.)
#endif

#include "sdl-freetype.h"
#include "sdl-freetype-utils.h"

static int
utf8_to_ucs4 (const char * str, FT_Int32 * ucs4, int len)
{
    const unsigned char * utf8 = (const unsigned char *)str;
    unsigned char ch;
    int extra;
    FT_Int32 unicode;

    if (!len)
        return 0;

    ch = *utf8++;

    if (!(ch & 0x80)) {
        unicode = ch;
        extra = 0;
    } else if (!(ch & 0x40)) {
        return -1;
    } else if (!(ch & 0x20)) {
        unicode = ch & 0x1f;
        extra = 1;
    } else if (!(ch & 0x10)) {
        unicode = ch & 0xf;
        extra = 2;
    } else if (!(ch & 0x08)) {
        unicode = ch & 0x07;
        extra = 3;
    } else if (!(ch & 0x04)) {
        unicode = ch & 0x03;
        extra = 4;
    } else if (!(ch & 0x02)) {
        unicode = ch & 0x01;
        extra = 5;
    } else {
        return -1;
    }

    if (extra > len - 1)
        return -1;

    while (extra--) {
        unicode <<= 6;
        ch = *utf8++;

        if ((ch & 0xc0) != 0x80)
            return -1;

        unicode |= ch & 0x3f;
    }

    *ucs4 = unicode;
    return utf8 - (const unsigned char *)str;
}

int
sdl_freetype_utf8_strlen (const char * utf8, int len)
{
    FT_Int32 ucs4;
    int clen;
    int ucs4len = 0;

    if (len < 0)
        len = strlen (utf8);

    while (len > 0) {
        clen = utf8_to_ucs4 (utf8, &ucs4, len);
        if (clen < 0)
            return -1;
        if (!ucs4)
            break;
        len -= clen;
        utf8 += clen;
        ucs4len++;
    }

    return ucs4len;
}

int
sdl_freetype_utf8_to_ucs4 (const char * utf8, int len, FT_Int32 ** retucs4)
{

    int i, slen, clen;
    FT_Int32 * ucs4;

    if (len < 0)
        len = strlen (utf8);

    slen = sdl_freetype_utf8_strlen (utf8, len);
    if (slen < 0)
        return -1;

    ucs4 = malloc (sizeof(FT_Int32) * (slen + 1));
    if (!ucs4)
        return -1;

    for (i = 0; i < slen; i++, len -= clen, utf8 += clen)
        clen = utf8_to_ucs4 (utf8, ucs4 + i, len);

    ucs4[slen] = 0;
    *retucs4 = ucs4;
    return slen;
}

#define UNI_SUR_HIGH_START 0xd800
#define UNI_SUR_HIGH_END   0xdbff
#define UNI_SUR_LOW_START  0xdc00
#define UNI_SUR_LOW_END    0xdfff
int
sdl_freetype_utf16_to_ucs4 (const short * utf16, int len, FT_Int32 ** retucs4)
{

    int i, clen;
    FT_Int32 * ucs4;

    if (len < 0) {
	const short * s = utf16;
	len = 0;

	while (*s++)
	    len++;
    }

    ucs4 = malloc (sizeof(FT_Int32) * (len + 1));
    if (!ucs4)
        return -1;

    for (i = 0; len > 0; i++) {
	unsigned int ch;

	ch = *utf16++ & 0xffff;
	len--;
	if (ch >= UNI_SUR_HIGH_START && ch <= UNI_SUR_HIGH_END &&
	    len > 0) {
	    unsigned int ch2 = *utf16 & 0xfffff;

	    if (ch2 >= UNI_SUR_LOW_START && ch2 <= UNI_SUR_LOW_END) {
		ch = (ch - UNI_SUR_HIGH_START) << 10;
		ch +=  ch2 - UNI_SUR_LOW_START + 0x0010000;
		utf16++;
		len--;
	    }
	}
	ucs4[i] = ch;
    }

    ucs4[i] = 0;
    *retucs4 = ucs4;
    return i;
}

/* code from cairo */
/* This function is identical to the C99 function lround(), except that it
 * performs arithmetic rounding (instead of away-from-zero rounding) and
 * has a valid input range of (INT_MIN, INT_MAX] instead of
 * [INT_MIN, INT_MAX]. It is much faster on both x86 and FPU-less systems
 * than other commonly used methods for rounding (lround, round, rint, lrint
 * or float (d + 0.5)).
 *
 * The reason why this function is much faster on x86 than other
 * methods is due to the fact that it avoids the fldcw instruction.
 * This instruction incurs a large performance penalty on modern Intel
 * processors due to how it prevents efficient instruction pipelining.
 *
 * The reason why this function is much faster on FPU-less systems is for
 * an entirely different reason. All common rounding methods involve multiple
 * floating-point operations. Each one of these operations has to be
 * emulated in software, which adds up to be a large performance penalty.
 * This function doesn't perform any floating-point calculations, and thus
 * avoids this penalty.
  */
int
sdl_freetype_lround (double d)
{
    uint32_t top, shift_amount, output;
    union {
        double d;
        uint64_t ui64;
        uint32_t ui32[2];
    } u;

    u.d = d;

    /* If the integer word order doesn't match the float word order, we swap
     * the words of the input double. This is needed because we will be
     * treating the whole double as a 64-bit unsigned integer. Notice that we
     * use WORDS_BIGENDIAN to detect the integer word order, which isn't
     * exactly correct because WORDS_BIGENDIAN refers to byte order, not word
     * order. Thus, we are making the assumption that the byte order is the
     * same as the integer word order which, on the modern machines that we
     * care about, is OK.
     */
#if ( defined(FLOAT_WORDS_BIGENDIAN) && !defined(WORDS_BIGENDIAN)) || \
    (!defined(FLOAT_WORDS_BIGENDIAN) &&  defined(WORDS_BIGENDIAN))
    {
        uint32_t temp = u.ui32[0];
        u.ui32[0] = u.ui32[1];
        u.ui32[1] = temp;
    }
#endif

#ifdef WORDS_BIGENDIAN
    #define MSW (0) /* Most Significant Word */
    #define LSW (1) /* Least Significant Word */
#else
    #define MSW (1)
    #define LSW (0)
#endif

    /* By shifting the most significant word of the input double to the
     * right 20 places, we get the very "top" of the double where the exponent
     * and sign bit lie.
     */
    top = u.ui32[MSW] >> 20;

    /* Here, we calculate how much we have to shift the mantissa to normalize
     * it to an integer value. We extract the exponent "top" by masking out the
     * sign bit, then we calculate the shift amount by subtracting the exponent
     * from the bias. Notice that the correct bias for 64-bit doubles is
     * actually 1075, but we use 1053 instead for two reasons:
     *
     *  1) To perform rounding later on, we will first need the target
     *     value in a 31.1 fixed-point format. Thus, the bias needs to be one
     *     less: (1075 - 1: 1074).
     *
     *  2) To avoid shifting the mantissa as a full 64-bit integer (which is
     *     costly on certain architectures), we break the shift into two parts.
     *     First, the upper and lower parts of the mantissa are shifted
     *     individually by a constant amount that all valid inputs will require
     *     at the very least. This amount is chosen to be 21, because this will
     *     allow the two parts of the mantissa to later be combined into a
     *     single 32-bit representation, on which the remainder of the shift
     *     will be performed. Thus, we decrease the bias by an additional 21:
     *     (1074 - 21: 1053).
     */
    shift_amount = 1053 - (top & 0x7FF);

    /* We are done with the exponent portion in "top", so here we shift it off
     * the end.
     */
    top >>= 11;

    /* Before we perform any operations on the mantissa, we need to OR in
     * the implicit 1 at the top (see the IEEE-754 spec). We needn't mask
     * off the sign bit nor the exponent bits because these higher bits won't
     * make a bit of difference in the rest of our calculations.
     */
    u.ui32[MSW] |= 0x100000;

    /* If the input double is negative, we have to decrease the mantissa
     * by a hair. This is an important part of performing arithmetic rounding,
     * as negative numbers must round towards positive infinity in the
     * halfwase case of -x.5. Since "top" contains only the sign bit at this
     * point, we can just decrease the mantissa by the value of "top".
     */
    u.ui64 -= top;

    /* By decrementing "top", we create a bitmask with a value of either
     * 0x0 (if the input was negative) or 0xFFFFFFFF (if the input was positive
     * and thus the unsigned subtraction underflowed) that we'll use later.
     */
    top--;

    /* Here, we shift the mantissa by the constant value as described above.
     * We can emulate a 64-bit shift right by 21 through shifting the top 32
     * bits left 11 places and ORing in the bottom 32 bits shifted 21 places
     * to the right. Both parts of the mantissa are now packed into a single
     * 32-bit integer. Although we severely truncate the lower part in the
     * process, we still have enough significant bits to perform the conversion
     * without error (for all valid inputs).
     */
    output = (u.ui32[MSW] << 11) | (u.ui32[LSW] >> 21);

    /* Next, we perform the shift that converts the X.Y fixed-point number
     * currently found in "output" to the desired 31.1 fixed-point format
     * needed for the following rounding step. It is important to consider
     * all possible values for "shift_amount" at this point:
     *
     * - {shift_amount < 0} Since shift_amount is an unsigned integer, it
     *   really can't have a value less than zero. But, if the shift_amount
     *   calculation above caused underflow (which would happen with
     *   input > INT_MAX or input <= INT_MIN) then shift_amount will now be
     *   a very large number, and so this shift will result in complete
     *   garbage. But that's OK, as the input was out of our range, so our
     *   output is undefined.
     *
     * - {shift_amount > 31} If the magnitude of the input was very small
     *   (i.e. |input| << 1.0), shift_amount will have a value greater than
     *   31. Thus, this shift will also result in garbage. After performing
     *   the shift, we will zero-out "output" if this is the case.
     *
     * - {0 <= shift_amount < 32} In this case, the shift will properly convert
     *   the mantissa into a 31.1 fixed-point number.
     */
    output >>= shift_amount;

    /* This is where we perform rounding with the 31.1 fixed-point number.
     * Since what we're after is arithmetic rounding, we simply add the single
     * fractional bit into the integer part of "output", and just keep the
     * integer part.
     */
    output = (output >> 1) + (output & 1);

    /* Here, we zero-out the result if the magnitude if the input was very small
     * (as explained in the section above). Notice that all input out of the
     * valid range is also caught by this condition, which means we produce 0
     * for all invalid input, which is a nice side effect.
     *
     * The most straightforward way to do this would be:
     *
     *      if (shift_amount > 31)
     *          output = 0;
     *
     * But we can use a little trick to avoid the potential branch. The
     * expression (shift_amount > 31) will be either 1 or 0, which when
     * decremented will be either 0x0 or 0xFFFFFFFF (unsigned underflow),
     * which can be used to conditionally mask away all the bits in "output"
     * (in the 0x0 case), effectively zeroing it out. Certain, compilers would
     * have done this for us automatically.
     */
    output &= ((shift_amount > 31) - 1);

    /* If the input double was a negative number, then we have to negate our
     * output. The most straightforward way to do this would be:
     *
     *      if (!top)
     *          output = -output;
     *
     * as "top" at this point is either 0x0 (if the input was negative) or
     * 0xFFFFFFFF (if the input was positive). But, we can use a trick to
     * avoid the branch. Observe that the following snippet of code has the
     * same effect as the reference snippet above:
     *
     *      if (!top)
     *          output = 0 - output;
     *      else
     *          output = output - 0;
     *
     * Armed with the bitmask found in "top", we can condense the two statements
     * into the following:
     *
     *      output = (output & top) - (output & ~top);
     *
     * where, in the case that the input double was negative, "top" will be 0,
     * and the statement will be equivalent to:
     *
     *      output = (0) - (output);
     *
     * and if the input double was positive, "top" will be 0xFFFFFFFF, and the
     * statement will be equivalent to:
     *
     *      output = (output) - (0);
     *
     * Which, as pointed out earlier, is equivalent to the original reference
     * snippet.
     */
    output = (output & top) - (output & ~top);

    return output;
#undef MSW
#undef LSW
}

/* This is the "magic number" approach to converting a double into fixed
 * point as described here:
 *
 * http://www.stereopsis.com/sree/fpu2006.html (an overview)
 * http://www.d6.com/users/checker/pdfs/gdmfp.pdf (in detail)
 *
 * The basic idea is to add a large enough number to the double that the
 * literal floating point is moved up to the extent that it forces the
 * double's value to be shifted down to the bottom of the mantissa (to make
 * room for the large number being added in). Since the mantissa is, at a
 * given moment in time, a fixed point integer itself, one can convert a
 * float to various fixed point representations by moving around the point
 * of a floating point number through arithmetic operations. This behavior
 * is reliable on most modern platforms as it is mandated by the IEEE-754
 * standard for floating point arithmetic.
 *
 * For our purposes, a "magic number" must be carefully selected that is
 * both large enough to produce the desired point-shifting effect, and also
 * has no lower bits in its representation that would interfere with our
 * value at the bottom of the mantissa. The magic number is calculated as
 * follows:
 *
 *          (2 ^ (MANTISSA_SIZE - FRACTIONAL_SIZE)) * 1.5
 *
 * where in our case:
 *  - MANTISSA_SIZE for 64-bit doubles is 52
 *  - FRACTIONAL_SIZE for 16.16 fixed point is 16
 *
 * Although this approach provides a very large speedup of this function
 * on a wide-array of systems, it does come with two caveats:
 *
 * 1) It uses banker's rounding as opposed to arithmetic rounding.
 * 2) It doesn't function properly if the FPU is in single-precision
 *    mode.
 */
#define SDL_FREETYPE_MAGIC_NUMBER_FIXED_16_16 (103079215104.0)
sdl_freetype_fixed_t
sdl_freetype_fixed_from_double (double d)
{
    union {
        double d;
        int32_t i[2];
    } u;

    u.d = d + SDL_FREETYPE_MAGIC_NUMBER_FIXED_16_16;
#ifdef FLOAT_WORDS_BIGENDIAN
    return u.i[1];
#else
    return u.i[0];
#endif
}

sdl_freetype_fixed_t
sdl_freetype_fixed_from_26_6 (FT_Int32 i)
{
    return i << 10;
}

sdl_freetype_fixed_t
sdl_freetype_fixed_from_integer (int f)
{
    return f << 16;
}

double
sdl_freetype_fixed_to_double (sdl_freetype_fixed_t f)
{
    return ((double) f) / 65536.0;
}

int
sdl_freetype_fixed_is_integer (sdl_freetype_fixed_t f)
{
    return (f & 0xFFFF) == 0;
}

int
sdl_freetype_fixed_integer (sdl_freetype_fixed_t f)
{
    return f >> 16;
}

int
sdl_freetype_fixed_round (sdl_freetype_fixed_t f)
{
    return (f + 32768)>> 16;
}

int
sdl_freetype_fixed_floor (sdl_freetype_fixed_t f)
{
    if (f >= 0)
        return f >> 16;
    else
        return -((-f - 1) >> 16) - 1;
}

int
sdl_freetype_fixed_ceil (sdl_freetype_fixed_t f)
{
    if (f > 0)
        return ((f - 1)>>16) + 1;
    else
        return - (-f >> 16);
}

void
sdl_freetype_text_extents_from_fixed (sdl_freetype_text_extents_t * extents,
                                      sdl_freetype_text_fixed_extents_t * fixed)
{
   extents->x_bearing = sdl_freetype_fixed_to_double(fixed->x_bearing);
   extents->y_bearing = sdl_freetype_fixed_to_double(fixed->y_bearing);
   extents->x_advance = sdl_freetype_fixed_to_double(fixed->x_advance);
   extents->y_advance = sdl_freetype_fixed_to_double(fixed->y_advance);
   extents->width = sdl_freetype_fixed_to_double(fixed->width);
   extents->height = sdl_freetype_fixed_to_double(fixed->height);
}
